import abc
import typing
import inspect
import dataclasses

import sklearn.base


class FeatureExtractorBase(sklearn.base.TransformerMixin, abc.ABC):
    """
    Base feature extractor class.

    Attributes
    ----------
    PRESETS : ``Presets``
        Settings presets of the feature extractor.
    """

    @dataclasses.dataclass(frozen=True)
    class Presets:
        """
        Settings presets container of the feature extractor.

        Attributes
        ----------
        full : ``object``
            The full, slow pipeline.
        reduced : ``object``
            The reduced pipeline with good balance between speed and quality.
        quick : ``object``
            The quick pipeline.
        """

        full: object
        reduced: object
        quick: object

    PRESETS: Presets = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if "settings" not in inspect.signature(cls.__init__).parameters.keys():
            return
        if cls.PRESETS is None:
            raise TypeError(f"{cls.__name__} must define PRESETS")
        if not isinstance(cls.PRESETS, FeatureExtractorBase.Presets):
            raise TypeError(f"{cls.__name__} must be an instance of Presets")

    def nest_feature_names(self, prefix: str, names: typing.List[str]) -> typing.List[str]:
        return [f"{prefix} -> {name}" for name in names]

    @abc.abstractmethod
    def feature_names(self) -> typing.List[str]:
        """
        Gives a list of features extracted by this class.

        Returns
        -------
        ``list[str]``
            Feature names.
        """
        pass



import abc
import typing

import numpy

import cvtda.utils
import cvtda.dumping
import cvtda.logging

from .. import utils


class Extractor(cvtda.utils.FeatureExtractorBase, abc.ABC):
    """
    Base class for transforming images into features.
    For grayscale images, directly computes the features.
    For rgb images, converts them to grayscale and recursively calls itself to compute the features.
    """

    def __init__(self, n_jobs: int = -1, only_get_from_dump: bool = False, **kwargs):
        """
        Parameters
        ----------
        n_jobs : ``int``, default: ``-1``
            The number of jobs to use for the computation. See :mod:`joblib` for details.
        only_get_from_dump : ``bool``, default `False`
            If true, all results will be obtained from dump, and no computations will be performed.
        """
        self.n_jobs_ = n_jobs
        self.only_get_from_dump_ = only_get_from_dump

        self.kwargs_ = kwargs
        self.kwargs_["n_jobs"] = n_jobs
        self.kwargs_["only_get_from_dump"] = only_get_from_dump

        self.fitted_ = False
        self.fit_dimensions_ = None

    def final_dump_name_(self, dump_name: typing.Optional[str] = None):
        """
        The name of the dump with the final result of the process.
        """
        return self.features_dump_(dump_name)

    def features_dump_(self, dump_name: typing.Optional[str]):
        """
        The name of the dump with the computed features.
        """
        return cvtda.dumping.dump_name_concat(dump_name, "features")

    def force_numpy_(self):
        """
        Whether to force numpy output from :meth:`transform`
        """
        return True

    def feature_names(self) -> typing.List[str]:
        if self.is_rgb_(self.fit_dimensions_):
            return [
                *self.nest_feature_names("rgb", self.feature_names_rgb_()),
                *self.nest_feature_names("gray", self.gray_extractor_.feature_names()),
                *self.nest_feature_names("red", self.red_extractor_.feature_names()),
                *self.nest_feature_names("green", self.green_extractor_.feature_names()),
                *self.nest_feature_names("blue", self.blue_extractor_.feature_names()),
            ]
        else:
            return self.feature_names_gray_()

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        self.fit_transform(images, dump_name)
        return self

    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        assert self.fitted_ is True, "fit() must be called before transform()"
        return self.process_(images, do_fit=False, dump_name=dump_name)

    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        result = self.process_(images, do_fit=True, dump_name=dump_name)
        self.fitted_ = True
        return result

    def is_rgb_(self, shape) -> bool:
        return (len(shape) == 3) and (shape[2] == 3)

    def process_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        """
        Processes a set of images: computes the features and, optionally, fits the extractor.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height x num_channels`) Input images in grayscale or rgb format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``numpy.ndarray`` or ``list[numpy.ndarray]``
            The result.
        """

        # Ensure `transform` got the same image format as `fit`.
        if self.fit_dimensions_ is not None:
            assert self.fit_dimensions_ == images.shape[1:], (
                f"The pipeline is fit for {self.fit_dimensions_}. Cannot use it with {images.shape}."
            )
        self.fit_dimensions_ = images.shape[1:]

        # Handle dumping.
        final_dump = self.final_dump_name_(dump_name)
        if self.only_get_from_dump_ and not self.is_rgb_(self.fit_dimensions_):
            return cvtda.dumping.dumper().get_dump(final_dump)
        if not do_fit and cvtda.dumping.dumper().has_dump(final_dump):
            return cvtda.dumping.dumper().get_dump(final_dump)

        if self.is_rgb_(self.fit_dimensions_):
            # For RGB, computes some features directly, and also transforms to four grayscale representations.
            cvtda.logging.logger().print("RGB images received. Transforming to grayscale.")

            rgb_dump = cvtda.dumping.dump_name_concat(dump_name, "rgb")
            gray_dump = cvtda.dumping.dump_name_concat(dump_name, "gray")
            red_dump = cvtda.dumping.dump_name_concat(dump_name, "red")
            green_dump = cvtda.dumping.dump_name_concat(dump_name, "green")
            blue_dump = cvtda.dumping.dump_name_concat(dump_name, "blue")

            if do_fit:
                # To process the grayscale representations, we clone the class for simplicity, so that
                # derived subclasses do not have to care about not mixing rgb and grayscale accidentally.
                self.gray_extractor_ = self.__class__(**self.kwargs_)
                self.red_extractor_ = self.__class__(**self.kwargs_)
                self.green_extractor_ = self.__class__(**self.kwargs_)
                self.blue_extractor_ = self.__class__(**self.kwargs_)

            # Handle RGB
            if self.only_get_from_dump_:
                rgb_data = cvtda.dumping.dumper().get_dump(self.final_dump_name_(rgb_dump))
            else:
                rgb_data = self.process_rgb_(images, do_fit, rgb_dump)

            # Handle grayscale
            gray_images = cvtda.utils.rgb2gray(images, self.n_jobs_)
            result = [
                rgb_data,
                utils.process_iter(self.gray_extractor_, gray_images, do_fit, gray_dump),
                utils.process_iter(self.red_extractor_, images[:, :, :, 0], do_fit, red_dump),
                utils.process_iter(self.green_extractor_, images[:, :, :, 1], do_fit, green_dump),
                utils.process_iter(self.blue_extractor_, images[:, :, :, 2], do_fit, blue_dump),
            ]
        else:
            result = [self.process_gray_(images, do_fit, dump_name)]
        return utils.hstack(result, self.force_numpy_())

    @abc.abstractmethod
    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        """
        Processes a set of RGB images.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height x num_channels`) Input images in rgb format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``numpy.ndarray`` or ``list[numpy.ndarray]``
            The result.
        """
        pass

    @abc.abstractmethod
    def feature_names_rgb_(self) -> typing.List[str]:
        """
        Gives a list of features extracted by this class for RGB images.

        Returns
        -------
        ``list[str]``
            Feature names.
        """
        pass

    @abc.abstractmethod
    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        """
        Processes a set of grayscale images.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height`) Input images in grayscale format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``numpy.ndarray`` or ``list[numpy.ndarray]``
            The result.
        """
        pass

    @abc.abstractmethod
    def feature_names_gray_(self) -> typing.List[str]:
        """
        Gives a list of features extracted by this class for grayscale images.

        Returns
        -------
        ``list[str]``
            Feature names.
        """
        pass



from .base import BaseLogger
from .cli import CLILogger
from .devnull import DevNullLogger

__all__ = [
    "BaseLogger",
    "CLILogger",
    "DevNullLogger",
]


def logger() -> BaseLogger:
    """
    Returns
    -------
    ``BaseLogger``
        Current active logger.
    """
    if BaseLogger.current_logger is None:
        BaseLogger.current_logger = CLILogger()
    return BaseLogger.current_logger



import typing

import numpy
import sklearn.base

import cvtda.dumping
from .process_iter import process_iter


def process_iter_dump(
    transformer: sklearn.base.TransformerMixin,
    data: numpy.ndarray,
    do_fit: bool,
    dump_name: typing.Optional[str] = None,
    *args,
    **kwargs,
):
    """
    Processes the `data` with the `transformer`, optionally fitting the transformer, and dumps the result.

    Parameters
    ----------
    transformer : ``sklearn.base.TransformerMixin``
        The transformer.
    data : ``numpy.ndarray``
        The data to process.
    do_fit : ``bool``
        If True, will also fit the transformer.
    dump_name : ``str``, optional
        The name of the dump for the results.

    Returns
    -------
    ``Any``
        The output of `transformer.transform`
    """
    if do_fit and cvtda.dumping.dumper().has_dump(dump_name):
        # We must call fit() anyway, even if we have a dump
        transformer.fit(data, *args, **kwargs)

    def func():
        return process_iter(transformer, data, do_fit, *args, **kwargs)

    return func() if dump_name is None else cvtda.dumping.dumper().execute(func, dump_name)



import typing

import numpy
import torch
import sklearn.base
import sklearn.metrics

import cvtda.utils
import cvtda.logging
import cvtda.neural_network


class MultiImageDataset:
    """
    A single structure representing the dataset of multiple images per object with topological features.

    Attributes
    ----------
    features : ``torch.Tensor``
        (size `num_items x num_features`) Topological features for each image.
    labels : ``torch.Tensor``
        (size `num_items x ...`) The target variable.
        May be of any shape, if the first dimension is the number of items.
    images : ``list[torch.Tensor]``
        (size `num_items x num_channels x width x height`) Sets of images.
    """

    def __init__(self, features, labels, images):
        self.labels = torch.tensor(labels, dtype=torch.long)
        self.features = torch.tensor(features, dtype=torch.float32)
        self.images = [torch.tensor(imgs, dtype=torch.float32).permute((0, 3, 1, 2)) for imgs in images]


class MultiImageClassifier(sklearn.base.ClassifierMixin):
    """
    A simple opinionated classifier based on neural networks that can work with multiple images simultaneously.
    """

    def __init__(
        self,
        n_jobs: int = -1,
        random_state: int = 42,
        device: torch.device = torch.device("cuda"),
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        n_epochs: int = 100,
        skip_images: bool = False,
        skip_features: bool = False,
    ):
        self.n_jobs_ = n_jobs
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs

        self.skip_images_ = skip_images
        self.skip_features_ = skip_features

    def fit(self, train: MultiImageDataset, val: typing.Optional[MultiImageDataset]):
        """
        Trains the model on the given dataset.

        Parameters
        ----------
        train : ``MultiImageDataset``
            Training dataset.
        val : ``MultiImageDataset``, optional
            Validation dataset. If specified, the quality metrics on this dataset will be printed after every epoch.
        """

        # Set seed
        cvtda.utils.set_random_seed(self.random_state_)

        # Initialize dataloader
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.labels, train.features, *train.images),
            batch_size=self.batch_size_,
            shuffle=True,
        )

        # Initialize model
        self.init_(*next(iter(train_dl)))

        for epoch in range(self.n_epochs_):
            sum_loss = 0

            self.model_list_.train()
            for labels, features, *images in cvtda.logging.logger().pbar(train_dl, desc=f"Epoch {epoch}"):
                self.optimizer_.zero_grad()
                pred = self.forward_(features, *images)  # Forward pass
                loss = torch.nn.functional.cross_entropy(pred, labels.to(self.device_), reduction="mean")  # Loss
                loss.backward()
                self.optimizer_.step()  # Optimization step
                sum_loss += loss.item()
            postfix = {"loss": sum_loss, "lr": self.optimizer_.param_groups[0]["lr"]}
            self.scheduler_.step()

            # Print metrics on validation, if needed
            if val is not None:
                val_proba = self.predict_proba_(val)
                val_pred = numpy.argmax(val_proba, axis=1)
                postfix["val_acc"] = sklearn.metrics.accuracy_score(val.labels, val_pred)

            cvtda.logging.logger().print(f"Epoch {epoch}:", postfix)

        self.fitted_ = True
        return self

    def predict_proba(self, dataset: MultiImageDataset) -> numpy.ndarray:
        """
        Calculate predictions for all images in the dataset.

        Parameters
        ----------
        dataset : ``MultiImageDataset``
            Dataset.

        Returns
        -------
        ``torch.Tensor``
            (size `num_items x num_classes`) Probabilities of each object belonging to each class.
        """
        assert self.fitted_ is True, "fit() must be called before predict_proba()"
        cvtda.utils.set_random_seed(self.random_state_)
        return self.predict_proba_(dataset)

    def init_(self, labels: torch.Tensor, features: torch.Tensor, *images: typing.List[torch.Tensor]):
        # Create model
        self.features_base_ = (
            cvtda.neural_network.NNBase(num_diagrams=0, skip_diagrams=True, skip_images=True, skip_features=False)
            .to(self.device_)
            .train()
        )

        self.images_bases_ = torch.nn.ModuleList(
            [
                cvtda.neural_network.NNBase(num_diagrams=0, skip_diagrams=True, skip_images=False, skip_features=True)
                .to(self.device_)
                .train()
                for _ in range(len(images))
            ]
        )

        self.model_ = (
            torch.nn.Sequential(
                torch.nn.Dropout(0.4),
                torch.nn.LazyLinear(256),
                torch.nn.BatchNorm1d(256),
                torch.nn.GELU(),
                torch.nn.Dropout(0.3),
                torch.nn.Linear(256, 128),
                torch.nn.BatchNorm1d(128),
                torch.nn.GELU(),
                torch.nn.Dropout(0.2),
                torch.nn.Linear(128, 64),
                torch.nn.BatchNorm1d(64),
                torch.nn.GELU(),
                torch.nn.Dropout(0.1),
                torch.nn.Linear(64, 32),
                torch.nn.BatchNorm1d(32),
                torch.nn.GELU(),
                torch.nn.Linear(32, len(torch.unique(labels))),
            )
            .to(self.device_)
            .train()
        )

        self.model_list_ = torch.nn.ModuleList([self.features_base_, self.images_bases_, self.model_])

        # Create optimizer and scheduler
        self.optimizer_ = torch.optim.AdamW(params=self.model_list_.parameters(), lr=self.learning_rate_)

        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001

        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        features_result = self.features_base_(None, features.to(self.device_))
        images_result = [self.images_bases_[i](images[i].to(self.device_), None) for i in range(len(images))]

        # Ensure everything is initialized correctly
        result = []
        if not self.skip_features_:
            result.append(features_result)
        if not self.skip_images_:
            result.extend(images_result)
        self.model_(torch.cat(result, dim=1))

        cvtda.logging.logger().print(f"Input to LazyLinear: {self.model_[1].in_features}")
        cvtda.logging.logger().print(f"Parameters: {sum(p.numel() for p in self.model_list_.parameters())}")

    def forward_(self, features: torch.Tensor, *images: typing.List[torch.Tensor]) -> torch.Tensor:
        result = []

        if not self.skip_features_:
            result.append(self.features_base_(None, features.to(self.device_)))

        if not self.skip_images_:
            result.extend([self.images_bases_[i](images[i].to(self.device_), None) for i in range(len(images))])

        return self.model_(torch.cat(result, dim=1))

    def predict_proba_(self, dataset: MultiImageDataset) -> numpy.ndarray:
        # Initialize dataloader
        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.features, *dataset.images),
            batch_size=self.batch_size_,
            shuffle=False,
        )

        # Make predictions
        y_pred_proba = []
        self.model_list_.eval()
        with torch.no_grad():
            for features, *images in dl:
                y_pred_proba.append(self.forward_(features, *images))
        y_pred_proba = torch.vstack(y_pred_proba)
        # Apply softmax to make probabilities from logits
        return torch.nn.functional.softmax(y_pred_proba, dim=1).cpu().numpy()



from .Autoencoder import Autoencoder
from .estimate_quality import estimate_quality
from .try_autoencoders import try_autoencoders

__all__ = [
    "Autoencoder",
    "estimate_quality",
    "try_autoencoders",
]



import torch


class Dataset:
    """
    A single structure representing the dataset for segmentation with topological features.

    Attributes
    ----------
    images : ``list[torch.Tensor]``
        (size `num_items x num_channels x width x height`) Sets of images.
    labels : ``torch.Tensor``
        (size `num_items x width x height`) Target segmentation masks.
    features : ``torch.Tensor``
        (size `num_items x num_features`) Topological features for each image.
    """

    def __init__(self, images, features, masks):
        # Save images in a torch-compatible format.
        self.images = torch.tensor(images, dtype=torch.float32)
        if len(self.images.shape) == 4:
            if self.images.shape[-1] == 3:
                self.images = self.images.permute((0, 3, 1, 2))
        else:
            assert len(self.images.shape) == 3
            self.images = self.images.unsqueeze(1)

        # Save everything else.
        self.masks = torch.tensor(masks, dtype=torch.float32).unsqueeze(1)
        self.features = torch.tensor(features, dtype=torch.float32)



import typing

import numpy
import torch
import torchvision
import torch.utils.data

import cvtda.utils
import cvtda.logging
import cvtda.neural_network

from .estimate_quality import estimate_quality


class Autoencoder:
    """
    A simple opinionated autoencoder implementation following [1].

    References
    ----------
    .. [1] Mark A. Kramer "Nonlinear principal component analysis using autoassociative neural networks"
            https://doi.org/10.1002/aic.690370209
    """

    def __init__(
        self,
        random_state: int = 42,
        device: torch.device = torch.device("cuda"),
        batch_size: int = 128,
        learning_rate: float = 1e-3,
        n_epochs: int = 20,
        latent_dim: int = 256,
        skip_diagrams: bool = False,
        skip_images: bool = False,
        skip_features: bool = False,
        base=torchvision.models.resnet34,
    ):
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs
        self.latent_dim_ = latent_dim

        self.skip_diagrams_ = skip_diagrams
        self.skip_images_ = skip_images
        self.skip_features_ = skip_features
        self.base_ = base

    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        """
        Trains the model on the given dataset.

        Parameters
        ----------
        train : ``cvtda.neural_network.Dataset``
            Training dataset.
        val : ``cvtda.neural_network.Dataset``, optional
            Validation dataset. If specified, the quality metrics on this dataset will be printed after every epoch.
        """

        # Set seed
        cvtda.utils.set_random_seed(self.random_state_)

        # Initialize dataloader
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, torch.arange(len(train))),
            batch_size=self.batch_size_,
            shuffle=True,
        )

        # Initialize model
        self.init_(*next(iter(train_dl)), train)

        for epoch in range(self.n_epochs_):
            sum_loss = 0

            self.model_list_.train()
            for images, features, diagram_idxs in cvtda.logging.logger().pbar(train_dl, desc=f"Epoch {epoch}"):
                self.optimizer_.zero_grad()
                pred = self.decode_(self.encode_(images, features, diagram_idxs, train))  # Forward pass
                loss = torch.nn.functional.mse_loss(pred, images.to(self.device_), reduction="mean")  # Loss
                loss.backward()
                self.optimizer_.step()  # Optimization step
                sum_loss += loss.item()
            postfix = {"loss": sum_loss, "lr": self.optimizer_.param_groups[0]["lr"]}
            self.scheduler_.step()

            # Print metrics on validation, if needed
            if val is not None:
                decoded = self.decode(self.encode(val))
                postfix = {**postfix, **estimate_quality(decoded, val.images.permute((0, 2, 3, 1)).cpu().numpy())}

            cvtda.logging.logger().print(f"Epoch {epoch}:", postfix)

        self.fitted_ = True
        return self

    def encode(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        """
        Runs the encoder on a dataset.

        Parameters
        ----------
        dataset : ``cvtda.neural_network.Dataset``
            Dataset.

        Returns
        -------
        ``torch.Tensor``
            (size `num_items x latent_dim`) Encoded images in the latent space.
        """

        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.images, dataset.features, torch.arange(len(dataset))),
            batch_size=self.batch_size_,
            shuffle=False,
        )

        encoded = []
        self.model_list_.eval()
        with torch.no_grad():
            for images, features, diagram_idxs in dl:
                encoded.append(self.encode_(images, features, diagram_idxs, dataset))
        return torch.vstack(encoded).cpu().numpy()

    def decode(self, latent_vectors: numpy.ndarray) -> numpy.ndarray:
        """
        Runs the decoder on a dataset.

        Parameters
        ----------
        latent_vectors : ``torch.Tensor``
            Encoded images in the latent space.

        Returns
        -------
        ``torch.Tensor``
            (size `num_items x num_channels x width x height`) Decoded images.
        """

        dl = torch.utils.data.DataLoader(latent_vectors, batch_size=self.batch_size_, shuffle=False)

        decoded = []
        self.model_list_.eval()
        with torch.no_grad():
            for vecs in dl:
                decoded.append(self.decode_(vecs))
        return torch.vstack(decoded).permute((0, 2, 3, 1)).squeeze().cpu().numpy()

    def init_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset,
    ):
        # Create encoder
        diagrams = [] if self.skip_diagrams_ else dataset.get_diagrams(diagram_idxs)
        self.encoder_base_ = (
            cvtda.neural_network.NNBase(
                num_diagrams=len(diagrams) // 2,
                skip_diagrams=self.skip_diagrams_,
                skip_images=self.skip_images_,
                skip_features=self.skip_features_,
                images_n_channels=images.shape[1],
                base=self.base_,
            )
            .to(self.device_)
            .train()
        )
        self.encoder_ = torch.nn.LazyLinear(self.latent_dim_).to(self.device_).train()

        # Create decoder
        w = images.shape[2] // 4
        h = images.shape[3] // 4
        self.decoder_ = (
            torch.nn.Sequential(
                torch.nn.InstanceNorm1d(self.latent_dim_),
                torch.nn.Linear(self.latent_dim_, 128 * w * h),
                torch.nn.InstanceNorm1d(128 * w * h),
                torch.nn.Unflatten(1, (128, w, h)),
                torch.nn.ReLU(),
                torch.nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, output_padding=1, padding=1),
                torch.nn.InstanceNorm2d(64),
                torch.nn.ReLU(),
                torch.nn.Conv2d(64, 64, kernel_size=3, padding=1),
                torch.nn.InstanceNorm2d(64),
                torch.nn.ReLU(),
                torch.nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, output_padding=1, padding=1),
                torch.nn.InstanceNorm2d(32),
                torch.nn.ReLU(),
                torch.nn.Conv2d(32, 16, kernel_size=3, padding=1),
                torch.nn.InstanceNorm2d(16),
                torch.nn.ReLU(),
                torch.nn.Conv2d(16, images.shape[1], kernel_size=3, padding=1),
                torch.nn.Sigmoid(),
            )
            .to(self.device_)
            .train()
        )

        self.model_list_ = torch.nn.ModuleList([self.encoder_base_, self.encoder_, self.decoder_])

        # Create optimizer and scheduler
        self.optimizer_ = torch.optim.AdamW(params=self.model_list_.parameters(), lr=self.learning_rate_)

        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001

        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        # Ensure everything is initialized correctly
        encoded = self.encode_(images, features, diagram_idxs, dataset)
        assert encoded.shape == (images.shape[0], self.latent_dim_), encoded.shape

        decoded = self.decode(encoded)
        assert decoded.shape == images.permute((0, 2, 3, 1)).squeeze().shape, f"{decoded.shape} != {images.shape}"

        cvtda.logging.logger().print(
            f"Encoder Base parameters: {sum(p.numel() for p in self.encoder_base_.parameters())}"
        )
        cvtda.logging.logger().print(f"Decoder parameters: {sum(p.numel() for p in self.decoder_.parameters())}")

    def encode_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset,
    ) -> torch.Tensor:
        images = images.to(self.device_)
        features = features.to(self.device_)
        if self.skip_diagrams_:
            return self.encoder_(self.encoder_base_(images, features))
        return self.encoder_(self.encoder_base_(images, features, *dataset.get_diagrams(diagram_idxs)))

    def decode_(self, latent_vector: torch.Tensor) -> torch.Tensor:
        return self.decoder_(latent_vector.to(self.device_))



import typing

import tqdm
import tqdm.contrib

from .base import BaseLogger

T = typing.TypeVar("T")


class CLILogger(BaseLogger):
    """
    A logger that prints to the standard output.
    """

    def __init__(self):
        pass

    def verbosity(self) -> int:
        return 2

    def print(self, data: T, *args) -> None:
        print(data, *args)

    def pbar(
        self, data: typing.Iterable[T], total: int = None, desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        return tqdm.tqdm(data, total=total, desc=desc)

    def zip(self, *iterables, desc: typing.Optional[str] = None):
        return tqdm.contrib.tzip(*iterables, desc=desc)

    def set_pbar_postfix(self, pbar: tqdm.tqdm, data: dict):
        pbar.set_postfix(data)



import typing
import dataclasses

import numpy
import skimage.measure
import skimage.feature

import cvtda.utils
import cvtda.logging

from . import utils
from .interface import Extractor


class GrayGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    """
    Extracts geometric features from grayscale images:
    - DAISY [1], see :func:`skimage.feature.daisy`
    - SIFT [2], see :class:`skimage.feature.SIFT`
    - ORB [3], see :class:`skimage.feature.ORB`
    - HOG [4], see :func:`skimage.feature.hog`
    - Basic local features, see :func:`skimage.feature.multiscale_basic_features`
    - Blur effect [5], see :func:`skimage.measure.blur_effect`
    - Moments ($M$): see :func:`skimage.measure.moments`
    - Centroid: $M_{10} / M{00}$, see :func:`skimage.measure.centroid`
    - Central moments: see :func:`skimage.measure.moments_central`
    - Hu's image moments: see :func:`skimage.measure.moments_hu`
    - Eigenvalues of the intertia vector, see :func:`skimage.measure.inertia_tensor_eigvals`
    - Shannon entropy: see :func:`skimage.measure.shannon_entropy`
    - Statistics of the local curvature for 256 uniform binarization thresholds:
        - Euler number: see :func:`skimage.measure.euler_number`
        - Perimeter: see :func:`skimage.measure.perimeter`
        - Area

    References
    ----------
    .. [1] Engin Tola, Vincent Lepetit, Pascal Fua
            "A fast local descriptor for dense matching."
            https://ieeexplore.ieee.org/document/4587673/
    .. [2] McConnell RK (1986)
            "Method of and apparatus for pattern recognition."
            https://patents.google.com/patent/US4567610A/en
    .. [3] Ethan Rublee, Vincent Rabaud, Kurt Konolige, Gary Bradski
            ORB: An efficient alternative to SIFT or SURF
            https://ieeexplore.ieee.org/document/6126544
    .. [4] D.G. Lowe "Object recognition from local scale-invariant features"
            https://ieeexplore.ieee.org/document/790410
    .. [5] Frederique Crete, Thierry Dolmiere, Patricia Ladret, Marina Nicolas
            "The blur effect: perception and estimation with a new no-reference perceptual blur metric"
            https://doi.org/10.1117/12.702790
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        daisy : ``bool``, default True
            Whether to compute DAISY features.
        sift : ``bool``, default True
            Whether to compute SIFT features.
        orb : ``bool``, default True
            Whether to compute ORB features.
        hog : ``bool``, default True
            Whether to compute HOG features.
        basic : ``bool``, default True
            Whether to compute basic local features.
        blur_effect : ``bool``, default True
            Whether to compute blur effect.
        moments : ``bool``, default True
            Whether to compute the moments.
        centroid : ``bool``, default True
            Whether to compute the centroid.
        moments_central : ``bool``, default True
            Whether to compute the central moments.
        moments_hu : ``bool``, default True
            Whether to compute Hu's image moments.
        inertia_tensor_eigvals : ``bool``, default True
            Whether to compute the eigenvalues of the intertia vector.
        shannon_entropy : ``bool``, default True
            Whether to compute the Shannon entropy.
        curvature : ``bool``, default True
            Whether to compute the local curvature features.
        """

        reduced_stats: bool = True

        daisy: bool = True
        sift: bool = True
        orb: bool = True
        hog: bool = True
        basic: bool = True
        blur_effect: bool = True
        moments: bool = True
        centroid: bool = True
        moments_central: bool = True
        moments_hu: bool = True
        inertia_tensor_eigvals: bool = True
        shannon_entropy: bool = True
        curvature: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings(curvature=False)
    )

    FEATURE_NAMES = [
        "daisy",
        "sift",
        "orb",
        "hog",
        "basic",
        "blur_effect",
        "centroid",
        "inertia_tensor_eigvals",
        "moments",
        "moments_central",
        "moments_hu",
        "shannon_entropy",
        "curvature",
    ]

    def __init__(self, n_jobs: int = -1, settings=Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.daisy,
            settings.sift,
            settings.orb,
            settings.hog,
            settings.basic,
            settings.blur_effect,
            settings.centroid,
            settings.inertia_tensor_eigvals,
            settings.moments,
            settings.moments_central,
            settings.moments_hu,
            settings.shannon_entropy,
            settings.curvature,
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        return self.feature_names_

    def fit(self, gray_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(
            self.features_mask, GrayGeometryExtractor.FEATURE_CALCULATORS, GrayGeometryExtractor.FEATURE_NAMES
        ):
            if is_needed:
                self.feature_names_.extend([f"{name}-{i}" for i in range(len(calc(self, gray_images[0])))])
        self.fitted_ = True
        return self

    def transform(self, gray_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before transform()"

        def process_one_(gray_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(gray_image)), 0)

        pbar = cvtda.logging.logger().pbar(gray_images, desc="GrayGeometryExtractor")
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs=self.n_jobs_))
        assert features.shape == (len(gray_images), len(self.feature_names()))
        return features

    def calc_raw_(self, gray_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        """
        Calculates all features for one grayscale image.

        Parameters
        ----------
        gray_image : ``numpy.ndarray``
            (size `width x height`) Input image.
        """
        return [
            calc(self, gray_image)
            for is_needed, calc in zip(self.features_mask, GrayGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]

    def calc_daisy_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        image_shape = max(*gray_image.shape)
        return skimage.feature.daisy(
            gray_image,
            step=(6 * image_shape // 32),
            radius=(12 * image_shape // 32),
            rings=5,
            histograms=5,
            orientations=8,
        ).flatten()

    def calc_sift_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        sift = skimage.feature.SIFT()
        try:
            sift.detect_and_extract(gray_image)
            sift_descriptors = sift.descriptors.transpose()
            if sift_descriptors.shape[1] == 0:
                raise "How is this possible?"
        except:  # noqa: E722
            # sift may throw errors for degenerate images. We substitute those with zeros.
            sift_descriptors = numpy.zeros((128, 1))
        return cvtda.utils.sequence2features(numpy.ma.array(sift_descriptors), reduced=self.reduced_stats_).flatten()

    def calc_orb_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        orb = skimage.feature.ORB()
        try:
            orb.detect_and_extract(gray_image)
            orb_descriptors = orb.descriptors.transpose()
            if orb_descriptors.shape[1] == 0:
                raise "How is this possible?"
        except:  # noqa: E722
            # orb may throw errors for degenerate images. We substitute those with zeros.
            orb_descriptors = numpy.zeros((256, 1))
        return cvtda.utils.sequence2features(numpy.ma.array(orb_descriptors), reduced=self.reduced_stats_).flatten()

    def calc_hog_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.feature.hog(gray_image, pixels_per_cell=(gray_image.shape[0] // 4, gray_image.shape[1] // 4))

    def calc_basic_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        basic_features = skimage.feature.multiscale_basic_features(gray_image).reshape((-1, 24))
        return cvtda.utils.sequence2features(
            numpy.ma.array(basic_features.transpose()), reduced=self.reduced_stats_
        ).flatten()

    def calc_blur_effect_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return numpy.array([skimage.measure.blur_effect(gray_image)])

    def calc_centroid_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.centroid(gray_image)

    def calc_inertia_tensor_eigvals_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(gray_image))
        except:  # noqa: E722
            # inertia tensor is undefined for degenerate images.
            return numpy.zeros((2))

    def calc_moments_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments(gray_image, order=9).flatten()

    def calc_moments_central_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments_central(gray_image, order=9).flatten()

    def calc_moments_hu_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        moments_central = skimage.measure.moments_central(gray_image, order=9)
        moments_normalized = skimage.measure.moments_normalized(moments_central)
        return skimage.measure.moments_hu(moments_normalized).flatten()

    def calc_shannon_entropy_(self, gray_image: numpy.ndarray):
        return numpy.array([skimage.measure.shannon_entropy(gray_image)])

    def calc_curvature_(self, gray_image: numpy.ndarray) -> numpy.ndarray:
        min_, max_ = gray_image.min(), gray_image.max()
        assert (min_ >= 0) and (max_ <= 1), f"Bad image format: should be [0, 1]; received [{min_}, {max_}]"

        euler_numbers, area, perimeter = [], [], []
        for threshold in range(256):
            bin = gray_image > threshold / 255.0
            euler_numbers.append(skimage.measure.euler_number(bin))
            area.append(bin.sum())
            perimeter.append(skimage.measure.perimeter(bin))
        series = numpy.array([euler_numbers, area, perimeter])
        series_diff = numpy.array([numpy.diff(euler_numbers), numpy.diff(area), numpy.diff(perimeter)])

        return numpy.concatenate(
            [
                cvtda.utils.sequence2features(series, reduced=self.reduced_stats_).flatten(),
                cvtda.utils.sequence2features(series_diff, reduced=self.reduced_stats_).flatten(),
            ]
        )

    FEATURE_CALCULATORS = [
        calc_daisy_,
        calc_sift_,
        calc_orb_,
        calc_hog_,
        calc_basic_,
        calc_blur_effect_,
        calc_centroid_,
        calc_inertia_tensor_eigvals_,
        calc_moments_,
        calc_moments_central_,
        calc_moments_hu_,
        calc_shannon_entropy_,
        calc_curvature_,
    ]


class RGBGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    """
    Extracts geometric features from RGB images:
    - HOG [1], see :func:`skimage.feature.hog`
    - Moments ($M$): see :func:`skimage.measure.moments`
    - Centroid: $M_{10} / M{00}$, see :func:`skimage.measure.centroid`
    - Eigenvalues of the intertia vector, see :func:`skimage.measure.inertia_tensor_eigvals`
    - Central moments: see :func:`skimage.measure.moments_central`
    - Pearson correlations between color channels, see :func:`skimage.measure.pearson_corr_coeff`

    References
    ----------
    .. [1] D.G. Lowe "Object recognition from local scale-invariant features"
            https://ieeexplore.ieee.org/document/790410
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        hog : ``bool``, default True
            Whether to compute HOG features.
        moments : ``bool``, default True
            Whether to compute the moments.
        centroid : ``bool``, default True
            Whether to compute the centroid.
        moments_central : ``bool``, default True
            Whether to compute the central moments.
        inertia_tensor_eigvals : ``bool``, default True
            Whether to compute the eigenvalues of the intertia vector.
        corr_coef : ``bool``, default True
            Whether to compute the correlations between color channels.
        """

        reduced_stats: bool = True

        hog: bool = True
        moments: bool = True
        centroid: bool = True
        moments_central: bool = True
        inertia_tensor_eigvals: bool = True
        corr_coef: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    FEATURE_NAMES = ["hog", "centroid", "inertia_tensor_eigvals", "moments", "moments_central", "corr_coef"]

    def __init__(self, n_jobs: int = -1, settings=Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.hog,
            settings.centroid,
            settings.inertia_tensor_eigvals,
            settings.moments,
            settings.moments_central,
            settings.corr_coef,
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        return self.feature_names_

    def fit(self, rgb_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(
            self.features_mask, RGBGeometryExtractor.FEATURE_CALCULATORS, RGBGeometryExtractor.FEATURE_NAMES
        ):
            if is_needed:
                self.feature_names_.extend([f"{name}-{i}" for i in range(len(calc(self, rgb_images[0])))])
        self.fitted_ = True
        return self

    def transform(self, rgb_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before transform()"

        def process_one_(rgb_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(rgb_image)), 0)

        pbar = cvtda.logging.logger().pbar(rgb_images, desc="RGBGeometryExtractor")
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs=self.n_jobs_))
        assert features.shape == (len(rgb_images), len(self.feature_names()))
        return features

    def calc_raw_(self, rgb_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        """
        Calculates all features for one RGB image.

        Parameters
        ----------
        rgb_image : ``numpy.ndarray``
            (size `width x height x 3`) Input image.
        """
        return [
            calc(self, rgb_image)
            for is_needed, calc in zip(self.features_mask, RGBGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]

    def calc_hog_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.feature.hog(rgb_image, channel_axis=2)

    def calc_centroid_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.centroid(rgb_image)

    def calc_inertia_tensor_eigvals_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(rgb_image))
        except:  # noqa: E722
            # inertia tensor is undefined for degenerate images.
            return numpy.zeros((3))

    def calc_moments_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments(rgb_image, order=6).flatten()

    def calc_moments_central_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return skimage.measure.moments_central(rgb_image, order=6).flatten()

    def calc_corr_coef_(self, rgb_image: numpy.ndarray) -> numpy.ndarray:
        return numpy.array(
            [
                skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 1])[0],
                skimage.measure.pearson_corr_coeff(rgb_image[:, :, 0], rgb_image[:, :, 2])[0],
                skimage.measure.pearson_corr_coeff(rgb_image[:, :, 1], rgb_image[:, :, 2])[0],
            ]
        )

    FEATURE_CALCULATORS = [
        calc_hog_,
        calc_centroid_,
        calc_inertia_tensor_eigvals_,
        calc_moments_,
        calc_moments_central_,
        calc_corr_coef_,
    ]


class MultidimensionalGeometryExtractor(cvtda.utils.FeatureExtractorBase):
    """
    Extracts geometric features from grayscale images of higher dimensions (4D, 5D, etc.):
    - Basic local features, see :func:`skimage.feature.multiscale_basic_features`
    - Blur effect [1], see :func:`skimage.measure.blur_effect`
    - Moments ($M$): see :func:`skimage.measure.moments`
    - Centroid: $M_{10} / M{00}$, see :func:`skimage.measure.centroid`
    - Central moments: see :func:`skimage.measure.moments_central`
    - Eigenvalues of the intertia vector, see :func:`skimage.measure.inertia_tensor_eigvals`

    References
    ----------
    .. [1] Frederique Crete, Thierry Dolmiere, Patricia Ladret, Marina Nicolas
            "The blur effect: perception and estimation with a new no-reference perceptual blur metric"
            https://doi.org/10.1117/12.702790
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        basic : ``bool``, default True
            Whether to compute basic local features.
        blur_effect : ``bool``, default True
            Whether to compute blur effect.
        moments : ``bool``, default True
            Whether to compute the moments.
        centroid : ``bool``, default True
            Whether to compute the centroid.
        moments_central : ``bool``, default True
            Whether to compute the central moments.
        inertia_tensor_eigvals : ``bool``, default True
            Whether to compute the eigenvalues of the intertia vector.
        """

        reduced_stats: bool = True

        basic: bool = True
        blur_effect: bool = True
        moments: bool = True
        centroid: bool = True
        moments_central: bool = True
        inertia_tensor_eigvals: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    FEATURE_NAMES = ["basic", "blur_effect", "centroid", "inertia_tensor_eigvals", "moments", "moments_central"]

    def __init__(self, n_jobs: int = -1, settings=Settings()):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.feature_names_ = []
        self.reduced_stats_ = settings.reduced_stats
        self.features_mask = [
            settings.basic,
            settings.blur_effect,
            settings.centroid,
            settings.inertia_tensor_eigvals,
            settings.moments,
            settings.moments_central,
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        return self.feature_names_

    def fit(self, nd_images: numpy.ndarray):
        self.feature_names_ = []
        for is_needed, calc, name in zip(
            self.features_mask,
            MultidimensionalGeometryExtractor.FEATURE_CALCULATORS,
            MultidimensionalGeometryExtractor.FEATURE_NAMES,
        ):
            if is_needed:
                self.feature_names_.extend([f"{name}-{i}" for i in range(len(calc(self, nd_images[0])))])
        self.fitted_ = True
        return self

    def transform(self, nd_images: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before transform()"

        def process_one_(nd_image: numpy.ndarray) -> numpy.ndarray:
            return numpy.nan_to_num(numpy.concatenate(self.calc_raw_(nd_image)), 0)

        pbar = cvtda.logging.logger().pbar(nd_images, desc="MultidimensionalGeometryExtractor")
        features = numpy.stack(cvtda.utils.parallel(process_one_, pbar, n_jobs=self.n_jobs_))
        assert features.shape == (len(nd_images), len(self.feature_names()))
        return features

    def calc_raw_(self, nd_image: numpy.ndarray) -> typing.List[numpy.ndarray]:
        """
        Calculates all features for one grayscale image.

        Parameters
        ----------
        nd_image : ``numpy.ndarray``
            (size `width x height x n_dimensions`) Input image.
        """
        return [
            calc(self, nd_image)
            for is_needed, calc in zip(self.features_mask, MultidimensionalGeometryExtractor.FEATURE_CALCULATORS)
            if is_needed
        ]

    def calc_basic_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        basic_features = skimage.feature.multiscale_basic_features(nd_image)
        basic_features = basic_features.reshape((-1, basic_features.shape[-1]))
        return cvtda.utils.sequence2features(
            numpy.ma.array(basic_features.transpose()), reduced=self.reduced_stats_
        ).flatten()

    def calc_blur_effect_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return numpy.array([skimage.measure.blur_effect(nd_image)])

    def calc_centroid_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.centroid(nd_image)

    def calc_inertia_tensor_eigvals_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        try:
            return numpy.array(skimage.measure.inertia_tensor_eigvals(nd_image))
        except:  # noqa: E722
            # inertia tensor is undefined for degenerate images.
            return numpy.zeros((len(nd_image.shape)))

    def calc_moments_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.moments(nd_image, order=9).flatten()

    def calc_moments_central_(self, nd_image: numpy.ndarray) -> typing.Optional[numpy.ndarray]:
        return skimage.measure.moments_central(nd_image, order=9).flatten()

    FEATURE_CALCULATORS = [
        calc_basic_,
        calc_blur_effect_,
        calc_centroid_,
        calc_inertia_tensor_eigvals_,
        calc_moments_,
        calc_moments_central_,
    ]


class GeometryExtractor(Extractor):
    """
    Extracts geometric features from images in any format, determines the format internally.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        rgb : ``RGBGeometryExtractor.Settings``
            Settings for the RGB extractor.
        gray : ``GrayGeometryExtractor.Settings``
            Settings for the grayscale extractor.
        multidimensional : ``MultidimensionalGeometryExtractor.Settings``
            Settings for the multidimensional extractor.
        """

        rgb: RGBGeometryExtractor.Settings = RGBGeometryExtractor.Settings()
        gray: GrayGeometryExtractor.Settings = GrayGeometryExtractor.Settings()
        multidimensional: MultidimensionalGeometryExtractor.Settings = MultidimensionalGeometryExtractor.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(
            rgb=RGBGeometryExtractor.PRESETS.full,
            gray=GrayGeometryExtractor.PRESETS.full,
            multidimensional=MultidimensionalGeometryExtractor.PRESETS.full,
        ),
        reduced=Settings(
            rgb=RGBGeometryExtractor.PRESETS.reduced,
            gray=GrayGeometryExtractor.PRESETS.reduced,
            multidimensional=MultidimensionalGeometryExtractor.PRESETS.reduced,
        ),
        quick=Settings(
            rgb=RGBGeometryExtractor.PRESETS.quick,
            gray=GrayGeometryExtractor.PRESETS.quick,
            multidimensional=MultidimensionalGeometryExtractor.PRESETS.quick,
        ),
    )

    def __init__(self, n_jobs: int = -1, settings=Settings(), only_get_from_dump: bool = False):
        super().__init__(n_jobs=n_jobs, settings=settings, only_get_from_dump=only_get_from_dump)

        self.rgb_extractor_ = RGBGeometryExtractor(n_jobs=self.n_jobs_, settings=settings.rgb)
        self.gray_extractor_ = GrayGeometryExtractor(n_jobs=self.n_jobs_, settings=settings.gray)
        self.multidimensional_extractor_ = MultidimensionalGeometryExtractor(
            n_jobs=self.n_jobs_, settings=settings.multidimensional
        )

    def process_rgb_(
        self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None
    ) -> numpy.ndarray:
        return utils.process_iter_dump(self.rgb_extractor_, rgb_images, do_fit, self.features_dump_(dump_name))

    def feature_names_rgb_(self) -> typing.List[str]:
        return self.rgb_extractor_.feature_names()

    def process_gray_(
        self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None
    ) -> numpy.ndarray:
        if len(gray_images.shape) == 3:
            return utils.process_iter_dump(self.gray_extractor_, gray_images, do_fit, self.features_dump_(dump_name))
        else:
            return utils.process_iter_dump(
                self.multidimensional_extractor_, gray_images, do_fit, self.features_dump_(dump_name)
            )

    def feature_names_gray_(self) -> typing.List[str]:
        if len(self.fit_dimensions_) == 2:
            return self.gray_extractor_.feature_names()
        else:
            return self.multidimensional_extractor_.feature_names()



import abc
import typing

import numpy
import itertools
import matplotlib.pyplot as plt

import cvtda.utils
import cvtda.logging
import cvtda.neural_network


class BaseLearner(abc.ABC):
    """
    Base class of a face recognition model.
    """

    def __init__(self, n_jobs: int = -1):
        self.n_jobs_ = n_jobs

    @abc.abstractmethod
    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        """
        Trains the model on the given dataset.

        Parameters
        ----------
        train : ``cvtda.neural_network.Dataset``
            Training dataset.
        val : ``cvtda.neural_network.Dataset``, optional
            Validation dataset. If specified, the quality metrics on this dataset will be printed after every epoch.
        """
        pass

    def estimate_quality(self, dataset: cvtda.neural_network.Dataset, ax: typing.Optional[plt.Axes] = None):
        """
        Estimates the quality of the model on a given dataset:
        creates a :mod:`matplotlib` figure in a given axis showing the
        distributions of distances between photos of same person and different people.

        Parameters
        ----------
        dataset : ``cvtda.neural_network.Dataset``
            Dataset.
        ax : ``matplotlib.pyplot.Axes``, optional, default None
            A matplotlib axis to draw the result in.
        """

        def calculate_distance_(i: int, j: int):
            return (i, j, self.calculate_distance_(i, j, dataset))

        # Calculate distances for all pairs of objects in parallel
        idxs = list(itertools.product(range(len(dataset)), range(len(dataset))))
        pbar = cvtda.logging.logger().pbar(idxs, desc="Calculating pairwise distances")
        distances_flat = cvtda.utils.parallel(calculate_distance_, pbar, n_jobs=self.n_jobs_)

        # Partition pairs of photos of the same person and different people
        correct_dists, incorrect_dists = {}, {}
        for i, j, distance in cvtda.logging.logger().pbar(distances_flat, desc="Analyzing distances"):
            label1, label2 = dataset.get_labels([i, j])
            if label1 == label2:
                correct_dists[(i, j)] = distance
            else:
                incorrect_dists[(i, j)] = distance
        correct_dists_values = list(correct_dists.values())
        incorrect_dists_values = list(incorrect_dists.values())

        # Draw the results
        if ax is not None:
            ax.set_ylim(0, 1)
            ax.get_yaxis().set_ticks([])
            ax.plot(correct_dists_values, numpy.ones_like(correct_dists_values) * 0.35, "x", label="Same person")
            ax.plot(
                incorrect_dists_values, numpy.ones_like(incorrect_dists_values) * 0.65, "x", label="Different people"
            )
        return correct_dists, incorrect_dists

    @abc.abstractmethod
    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset) -> float:
        """
        Calculate the latent distance between two objects in the dataset.

        Parameters
        ----------
        first : ``int``
            Index of the first object in the dataset.
        second : ``int``
            Index of the second object in the dataset.
        dataset : ``cvtda.neural_network.Dataset``
            Dataset.

        Returns
        -------
        ``float``
            Distance between two objects in the latent space.
        """
        pass



import typing

import torch
import torchvision
import torch.utils.data
import pytorch_metric_learning.miners
import pytorch_metric_learning.losses
import pytorch_metric_learning.samplers
import pytorch_metric_learning.utils.accuracy_calculator

import cvtda.utils
import cvtda.logging
import cvtda.neural_network

from .BaseLearner import BaseLearner


class NNLearner(BaseLearner):
    """
    A neural network-based face recognition model trained for the triplet loss [1].

    References
    ----------
    .. [1] Hideki Oki, Motoshi Abe, Junichi Miyao, Takio Kurita
            "Triplet Loss for Knowledge Distillation"
            https://doi.org/10.48550/arXiv.2004.08116
    """

    def __init__(
        self,
        n_jobs: int = -1,
        random_state: int = 42,
        device: torch.device = torch.device("cuda"),
        batch_size: int = 32,
        learning_rate: float = 1e-3,
        n_epochs: int = 25,
        length_before_new_iter: typing.Optional[int] = None,
        margin: int = 0.1,
        latent_dim: int = 256,
        skip_diagrams: bool = False,
        skip_images: bool = False,
        skip_features: bool = False,
        base=torchvision.models.resnet34,
    ):
        super().__init__(n_jobs)
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs
        self.length_before_new_iter_ = length_before_new_iter or (self.batch_size_ * 20)

        self.margin_ = margin
        self.latent_dim_ = latent_dim

        self.skip_diagrams_ = skip_diagrams
        self.skip_images_ = skip_images
        self.skip_features_ = skip_features
        self.base_ = base

    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        # Set seed
        cvtda.utils.set_random_seed(self.random_state_)

        # Initialize sampler and dataloader
        train_mpc_sampler = pytorch_metric_learning.samplers.MPerClassSampler(
            m=4, labels=train.labels, length_before_new_iter=self.length_before_new_iter_
        )
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, train.labels, torch.arange(len(train))),
            batch_size=self.batch_size_,
            sampler=train_mpc_sampler,
        )

        # Initialize model
        self.init_(*next(iter(train_dl)), train)

        # Initialize triplet loss and relevant utilities
        train_miner = pytorch_metric_learning.miners.TripletMarginMiner(margin=self.margin_, type_of_triplets="all")
        train_loss = pytorch_metric_learning.losses.TripletMarginLoss(margin=self.margin_)
        metrics = pytorch_metric_learning.utils.accuracy_calculator.AccuracyCalculator()

        # Initialize validation dataloader
        if val is not None:
            val_dl = torch.utils.data.DataLoader(
                torch.utils.data.TensorDataset(val.images, val.features, val.labels, torch.arange(len(val))),
                batch_size=self.batch_size_,
            )

        pbar = cvtda.logging.logger().pbar(range(self.n_epochs_), desc="Train")
        for _ in pbar:
            sum_loss = 0

            self.model_list_.train()
            for images, features, labels, diagram_idxs in train_dl:
                self.optimizer_.zero_grad()
                embeddings = self.forward_(images, features, diagram_idxs, train)  # Forward pass
                indices = train_miner(embeddings, labels)
                loss = train_loss(embeddings, labels, indices)  # Loss
                loss.backward()
                self.optimizer_.step()  # Optimization step
                sum_loss += loss.item()
            postfix = {"loss": sum_loss}
            self.scheduler_.step()

            # Print metrics on validation, if needed
            if val is not None:
                self.model_list_.eval()
                all_embeddings, all_targets = [], []
                for images, features, labels, diagram_idxs in val_dl:
                    with torch.no_grad():
                        all_targets.append(labels)
                        all_embeddings.append(self.forward_(images, features, diagram_idxs, val))
                result = metrics.get_accuracy(torch.cat(all_embeddings, dim=0), torch.cat(all_targets, dim=0))
                postfix = {**postfix, **result}

            cvtda.logging.logger().set_pbar_postfix(pbar, postfix)

        self.model_list_.eval()
        self.fitted_ = True
        return self

    def calculate_distance_(self, first: int, second: int, dataset: cvtda.neural_network.Dataset):
        self.model_list_.eval()
        idx = [first, second]
        embeddings = self.forward_(dataset.images[idx], dataset.features[idx], idx, dataset)
        return torch.sqrt(torch.sum((embeddings[0] - embeddings[1]) ** 2)).item()

    def init_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        labels: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset,
    ):
        diagrams = [] if self.skip_diagrams_ else dataset.get_diagrams(diagram_idxs)

        is_no_topology = self.skip_diagrams_ and self.skip_features_ and not self.skip_images_
        cvtda.logging.logger().print("Topology: ", not is_no_topology)
        cvtda.logging.logger().print("Images: ", images.shape)
        cvtda.logging.logger().print("Features: ", features.shape)

        if is_no_topology:
            images_output = len(torch.unique(labels))
        else:
            images_output = self.latent_dim_

        # Create model
        self.nn_base_ = (
            cvtda.neural_network.NNBase(
                num_diagrams=len(diagrams) // 2,
                skip_diagrams=self.skip_diagrams_,
                skip_images=self.skip_images_,
                skip_features=self.skip_features_,
                images_n_channels=images.shape[1],
                images_output=images_output,
                base=self.base_,
            )
            .to(self.device_)
            .train()
        )

        self.model_ = (
            torch.nn.Sequential(
                torch.nn.Dropout(0.3),
                torch.nn.LazyLinear(1024),
                torch.nn.BatchNorm1d(1024),
                torch.nn.GELU(),
                torch.nn.Dropout(0.2),
                torch.nn.Linear(1024, 768),
                torch.nn.BatchNorm1d(768),
                torch.nn.GELU(),
                torch.nn.Dropout(0.1),
                torch.nn.Linear(768, 512),
                torch.nn.BatchNorm1d(512),
                torch.nn.GELU(),
                torch.nn.Linear(512, self.latent_dim_),
            )
            .to(self.device_)
            .train()
        )

        if is_no_topology:
            self.model_ = torch.nn.Identity()

        self.model_list_ = torch.nn.ModuleList([self.nn_base_, self.model_])

        # Create optimizer and scheduler
        self.optimizer_ = torch.optim.AdamW(params=self.model_list_.parameters(), lr=self.learning_rate_)

        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001

        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        # Ensure everything is initialized correctly
        self.forward_(images, features, diagram_idxs, dataset)
        if not is_no_topology:
            cvtda.logging.logger().print(f"Input to LazyLinear: {self.model_[1].in_features}")
        cvtda.logging.logger().print(f"Parameters: {sum(p.numel() for p in self.model_list_.parameters())}")

    def forward_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset,
    ) -> torch.Tensor:
        images = images.to(self.device_)
        features = features.to(self.device_)
        if self.skip_diagrams_:
            return self.model_(self.nn_base_(images, features))
        return self.model_(self.nn_base_(images, features, *dataset.get_diagrams(diagram_idxs)))



from .Dataset import Dataset
from .estimate_quality import estimate_quality
from .MiniUnet import MiniUnet
from .segment import segment

__all__ = [
    "Dataset",
    "estimate_quality",
    "MiniUnet",
    "segment",
]



import numpy


def spread_points(size: int, num_points: int) -> numpy.ndarray:
    """
    Spread a given number of points evenly on a line of given size.

    Parameters
    ----------
    size : ``int``
        Size of the line. Must be an odd number.
    num_points : ``int``
        Number of points to spread.

    Returns
    -------
    ``numpy.ndarray``
        (size `num_points`) The coordinates of points.
    """
    assert size % 2 == 0, "Even size is not implemented"

    center, spacing = size // 2, size // (num_points + 1)
    centers = numpy.array(range((spacing + 1) // 2, center, spacing + 1))
    centers = numpy.array([*(center - centers - 1), *(center + centers)])
    if num_points % 2 == 1:
        centers = numpy.array([*centers, center])
    centers.sort()
    return centers



import numpy
import sklearn.metrics


def estimate_quality(pred_proba: numpy.ndarray, masks: numpy.ndarray) -> dict:
    """
    Quality metrics for segmentation task.

    Parameters
    ----------
    pred_proba : ``numpy.ndarray``
        (size `num_items x width x height`) Predicted probability estimates for each pixel.
    masks : ``numpy.ndarray``
        (size `num_items x width x height`) Target segmentation masks.

    Returns
    -------
    ``dict``
        1. Accuracy. See :func:`sklearn.metrics.accuracy_score`.
        2. AUC-ROC. See :func:`sklearn.metrics.roc_auc_score`.
        3. Precision. See :func:`sklearn.metrics.precision_score`.
        4. Recall. See :func:`sklearn.metrics.recall_score`.
        5. F1-score. See :func:`sklearn.metrics.f1_score`.
        6. Jaccard score. See :func:`sklearn.metrics.jaccard_score`.
    """

    # Segmentation is basically just per-pixel classification
    scores = pred_proba.flatten()
    preds = (scores >= 0.5).astype(numpy.int8)
    targets = masks.flatten().astype(numpy.int8)
    return {
        "Accuracy": sklearn.metrics.accuracy_score(targets, preds),
        "AUC-ROC": sklearn.metrics.roc_auc_score(targets, scores),
        "Precision": sklearn.metrics.precision_score(targets, preds, zero_division=0),
        "Recall": sklearn.metrics.recall_score(targets, preds, zero_division=0),
        "F1-score": sklearn.metrics.f1_score(targets, preds, zero_division=0),
        "Jaccard score": sklearn.metrics.jaccard_score(targets, preds, zero_division=0),
    }



from .DiagramsLearner import DiagramsLearner
from .learn import learn
from .NNLearner import NNLearner
from .SimpleTopologicalLearner import SimpleTopologicalLearner

__all__ = [
    "DiagramsLearner",
    "learn",
    "NNLearner",
    "SimpleTopologicalLearner",
]



from .base import BaseDumper
from .numpy import NumpyDumper
from .devnull import DevNullDumper
from .dump_name_concat import dump_name_concat

__all__ = [
    "BaseDumper",
    "NumpyDumper",
    "DevNullDumper",
    "dump_name_concat",
]


def dumper() -> BaseDumper:
    """
    Returns
    -------
    ``BaseLogger``
        Current active dumper.
    """
    if BaseDumper.current_dumper is None:
        BaseDumper.current_dumper = NumpyDumper("./")
    return BaseDumper.current_dumper



import numpy
import numpy.ma.core


def base_(data: numpy.ndarray) -> numpy.ndarray:
    if data is numpy.ma.core.MaskedArray:
        return numpy.ma.hstack(
            [
                numpy.ma.sum(numpy.ma.abs(data), axis=1, keepdims=True),  # manhattan norm
                numpy.ma.sqrt(numpy.ma.sum(data**2, axis=1, keepdims=True)),  # euclidean norm
                numpy.ma.max(numpy.ma.abs(data), axis=1, keepdims=True),  # infinity norm
            ]
        ).filled(0)
    else:
        return numpy.hstack(
            [
                numpy.trapz(data, axis=1).reshape(-1, 1),  # integral
                numpy.sum(numpy.abs(data), axis=1, keepdims=True),  # manhattan norm
                numpy.sqrt(numpy.sum(data**2, axis=1, keepdims=True)),  # euclidean norm
                numpy.max(numpy.abs(data), axis=1, keepdims=True),  # infinity norm
            ]
        )


def extension_(data: numpy.ndarray) -> numpy.ndarray:
    if data is numpy.ma.core.MaskedArray:
        return numpy.ma.hstack(
            [
                numpy.ma.max(data, axis=1, keepdims=True),
                numpy.ma.sum(data, axis=1, keepdims=True),
                numpy.ma.mean(data, axis=1, keepdims=True),
                numpy.ma.std(data, axis=1, keepdims=True),
                numpy.ma.median(data, axis=1, keepdims=True),
            ]
        ).filled(0)
    else:
        return numpy.hstack(
            [
                numpy.max(data, axis=1, keepdims=True),
                numpy.sum(data, axis=1, keepdims=True),
                numpy.mean(data, axis=1, keepdims=True),
                numpy.std(data, axis=1, keepdims=True),
                numpy.median(data, axis=1, keepdims=True),
            ]
        )


def sequence2features(sequence_batch: numpy.ndarray, reduced: bool = True) -> numpy.ndarray:
    """
    Calculates statistical features of a numeric sequence for all objects in a batch:
    - Integral
    - Manhattan norm
    - Euclidean norm
    - Infinity norm
    - Maximum
    - Sum of elements
    - Mean of elements
    - Standard deviation
    - Median

    Parameters
    ----------
    sequence_batch : ``numpy.ndarray``
        (size `num_items x seq_length`) Input data: a batch of numeric sequences.
        Masked array are supported.
    reduced : ``bool``, default: ``True``
        Removes some features from computations to improve efficiency in reduced pipelines.

    Returns
    -------
    ``numpy.ndarray``
        (size `num_items x num_features`) The computed features.
    """
    base = base_(sequence_batch)
    if reduced:
        return base

    if sequence_batch is numpy.ma.core.MaskedArray:
        return numpy.ma.hstack([base, extension_(sequence_batch)])
    else:
        return numpy.hstack([base, extension_(sequence_batch)])



import math
import typing
import dataclasses

import numpy
import joblib
import sklearn.base
import gtda.diagrams

import cvtda.utils
import cvtda.logging


def determine_filtering_epsilon_(diagrams: numpy.ndarray, percentile: float) -> float:
    """
    Computes the absolute value of the threshold for filtering persistence diagrams.

    Parameters
    ----------
    diagrams : ``numpy.ndarray``
        (size `num_diagrams x num_points x 3`) Persistence diagrams.
    percentile : ``float``
        The fraction of least persistent features to remove with filtration.

    Returns
    -------
    ``float``
        Absolute value threshold for :class:`gtda.diagrams.Filtering`.
    """

    life = (diagrams[:, :, 1] - diagrams[:, :, 0]).flatten()
    if len(numpy.unique(life)) == 1:
        return 1e-8
    return numpy.percentile(life[life != 0], percentile)


class Vectorizer(cvtda.utils.FeatureExtractorBase):
    """
    Base class for a process that transforms persistence diagrams into feature vectors.
    """

    def __init__(self):
        self.feature_names_ = []
        self.fitted_ = False

    def fit(self, diagrams: numpy.ndarray):
        self.fitted_ = True
        self.homology_dimensions_ = numpy.unique(diagrams[:, :, 2])
        num_features = self.transform(diagrams[:32]).shape[1]
        self.feature_names_ = [f"{i}" for i in range(num_features)]
        return self

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before feature_names()"

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        return self.feature_names_


class SequenceStats(Vectorizer):
    """
    Base class for a process that transforms persistence diagrams into feature vectors
    by calculating statistical features of some measure for every homology dimension on the diagram.
    """

    def __init__(
        self, impl: typing.List[sklearn.base.TransformerMixin], enabled: bool = True, reduced_stats: bool = True
    ):
        super().__init__()
        self.impl_ = impl if enabled else []
        self.reduced_stats_ = reduced_stats

    def fit(self, diagrams: numpy.ndarray):
        for item in self.impl_:
            item.fit(diagrams)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        super().transform(diagrams)
        if len(self.impl_) == 0:
            return numpy.empty((len(diagrams), 0))
        flat_shape = (len(diagrams), len(self.homology_dimensions_), -1)
        return numpy.hstack([self.calc_stats_(item.transform(diagrams).reshape(flat_shape)) for item in self.impl_])

    def calc_stats_(self, data: numpy.ndarray):
        return numpy.hstack(
            [
                cvtda.utils.sequence2features(data[:, dim, :], reduced=self.reduced_stats_)
                for dim in range(data.shape[1])
            ]
        )


class Proxy(Vectorizer):
    """
    A cvtda-compatible proxy-wrapper for giotto-tda transformers.
    """

    def __init__(self, impl: sklearn.base.TransformerMixin, enabled: bool = True):
        self.impl_ = impl if enabled else None

    def fit(self, diagrams: numpy.ndarray):
        if self.impl_ is not None:
            self.impl_.fit(diagrams)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        super().transform(diagrams)
        if self.impl_ is not None:
            return self.impl_.transform(diagrams)
        return numpy.empty((len(diagrams), 0))


class BettiCurve(SequenceStats):
    """
    Computes statistical features of betti curves for persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.BettiCurve` for details.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        n_bins : ``int``, default = 64
            The number of filtration parameter values, per available homology
            dimension, to sample for discretization.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        """

        enabled: bool = True
        n_bins: int = 64
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    def __init__(self, settings=Settings()):
        super().__init__(
            [gtda.diagrams.BettiCurve(n_bins=settings.n_bins, n_jobs=1)], settings.enabled, settings.reduced_stats
        )


class Landscape(SequenceStats):
    """
    Computes statistical features of persistence landscapes [1] for persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.PersistenceLandscape` for details.

    References
    ----------
    .. [1] Peter Bubenik "Statistical topological data analysis using persistence landscapes"
            https://doi.org/10.48550/arXiv.1207.6437
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        n_bins : ``int``, default = 64
            The number of filtration parameter values, per available homology
            dimension, to sample for discretization.
        layers : ``int``, default = 3
            How many layers to consider in the persistence landscape.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        """

        enabled: bool = True
        n_bins: int = 64
        layers: int = 3
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    def __init__(self, settings=Settings()):
        impl = gtda.diagrams.PersistenceLandscape(n_layers=settings.layers, n_bins=settings.n_bins, n_jobs=1)
        super().__init__([impl], settings.enabled, settings.reduced_stats)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        if len(self.impl_) == 0:
            return numpy.empty((len(diagrams), 0))
        impl: gtda.diagrams.PersistenceLandscape = self.impl_[0]
        landscape = impl.transform(diagrams)
        return numpy.hstack(
            [self.calc_stats_(landscape[:, layer :: impl.n_layers, :]) for layer in range(impl.n_layers)]
        )


class Silhouette(SequenceStats):
    """
    Computes statistical features of persistence silhouettes [1] for persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.Silhouette` for details.

    References
    ----------
    .. [1] Frédéric Chazal, Brittany Terese Fasy, Fabrizio Lecci, Alessandro Rinaldo, Larry Wasserman
            "Stochastic Convergence of Persistence Landscapes and Silhouettes"
            https://doi.org/10.48550/arXiv.1312.0308
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        n_bins : ``int``, default = 64
            The number of filtration parameter values, per available homology
            dimension, to sample for discretization.
        powers : ``list[int]``, default = [1, 2]
            The powers to which persistence values are raised to define power-weighted silhouettes.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        """

        enabled: bool = True
        n_bins: int = 64
        powers: typing.List[int] = dataclasses.field(default_factory=lambda: [1, 2])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    def __init__(self, settings=Settings()):
        impl = [gtda.diagrams.Silhouette(power=power, n_bins=settings.n_bins, n_jobs=1) for power in settings.powers]
        super().__init__(impl, settings.enabled, settings.reduced_stats)


class HeatKernel(SequenceStats):
    """
    Computes statistical features of heat kernels applied over diagrams.
    See :class:`gtda.diagrams.HeatKernel` for details.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        n_bins : ``int``, default = 64
            The number of filtration parameter values, per available homology
            dimension, to sample for discretization.
        sigmas : ``list[int]``, default = [0.1, 1.0, numpy.pi]
            Standard deviation for Gaussian kernels.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        """

        enabled: bool = True
        n_bins: int = 64
        sigmas: typing.List[float] = dataclasses.field(default_factory=lambda: [0.1, 1.0, numpy.pi])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(enabled=False), quick=Settings(enabled=False)
    )

    def __init__(self, settings=Settings()):
        impl = [gtda.diagrams.HeatKernel(sigma=sigma, n_bins=settings.n_bins, n_jobs=1) for sigma in settings.sigmas]
        super().__init__(impl, settings.enabled, settings.reduced_stats)


class PersistenceImage(SequenceStats):
    """
    Computes statistical features of persistence images [1] for persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.Silhouette` for details.

    References
    ----------
    .. [1] Henry Adams, Sofya Chepushtanova, Tegan Emerson, Eric Hanson, Michael Kirby, Francis Motta,
            Rachel Neville, Chris Peterson, Patrick Shipman, Lori Ziegelmeier
            "Persistence Images: A Stable Vector Representation of Persistent Homology"
            https://doi.org/10.48550/arXiv.1507.06217
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        n_bins : ``int``, default = 64
            The number of filtration parameter values, per available homology
            dimension, to sample for discretization.
        sigmas : ``list[int]``, default = [0.1, 1.0, numpy.pi]
            Standard deviation for Gaussian kernels.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        """

        enabled: bool = True
        n_bins: int = 64
        sigmas: typing.List[float] = dataclasses.field(default_factory=lambda: [0.1, 1.0, numpy.pi])
        reduced_stats: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(enabled=False), quick=Settings(enabled=False)
    )

    def __init__(self, settings=Settings()):
        impl = [
            gtda.diagrams.PersistenceImage(sigma=sigma, n_bins=settings.n_bins, n_jobs=1) for sigma in settings.sigmas
        ]
        super().__init__(impl, settings.enabled, settings.reduced_stats)


class Entropy(Proxy):
    """
    Computes persistence entropy [1] for persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.PersistenceEntropy` for details.

    References
    ----------
    .. [1] Rucco Matteo, Castiglione Filippo, Merelli Emanuela, Pettini Marco
            "Characterisation of the Idiotypic Immune Network Through Persistent Entropy"
            https://doi.org/10.1007/978-3-319-29228-1_11
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        """

        enabled: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(full=Settings(), reduced=Settings(), quick=Settings())

    def __init__(self, settings=Settings()):
        super().__init__(gtda.diagrams.PersistenceEntropy(nan_fill_value=0, n_jobs=1), settings.enabled)


class NumberOfPoints(Proxy):
    """
    Computes the numbers of points on persistence diagrams per homology dimension.
    See :class:`gtda.diagrams.NumberOfPoints` for details.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        """

        enabled: bool = True

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(full=Settings(), reduced=Settings(), quick=Settings())

    def __init__(self, settings=Settings()):
        super().__init__(gtda.diagrams.NumberOfPoints(n_jobs=1), settings.enabled)


class Lifetime(SequenceStats):
    """
    Computes statistical features of simplex lifetimes for persistence diagrams per homology dimension:
    - Persistence (lifetime) of the simplex: $d - b$
    - Midlife point of the simplex lifetime: $(b + d) / 2$
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        enabled : ``bool``, default True
            Whether the extractor is enabled in the pipeline.
        reduced_stats : ``bool``, default True
            Whether to reduce the set of statistical features. See :func:`utils.sequence2features` for details.
        filtering_percentile : ``float``, default = 10
            The fraction of least persistent features to remove during filtration.
        """

        enabled: bool = True
        reduced_stats: bool = True
        filtering_percentile: float = 10

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(reduced_stats=False), reduced=Settings(), quick=Settings()
    )

    def __init__(self, settings=Settings()):
        super().__init__([], settings.enabled, settings.reduced_stats)
        self.filtering_percentile_ = settings.filtering_percentile
        self.enabled_ = settings.enabled

    def fit(self, diagrams: numpy.ndarray):
        self.filtering_epsilon_ = determine_filtering_epsilon_(diagrams, self.filtering_percentile_)
        return super().fit(diagrams)

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        Vectorizer.transform(self, diagrams)
        if not self.enabled_:
            return numpy.empty((len(diagrams), 0))

        birth, death, dim = diagrams[:, :, 0], diagrams[:, :, 1], diagrams[:, :, 2]
        bd2, life = (birth + death) / 2.0, death - birth
        bd2_bulk, life_bulk = [], []
        for d in self.homology_dimensions_:
            mask = (dim != d) | (life < self.filtering_epsilon_)
            bd2_bulk.append(numpy.ma.array(bd2, mask=mask))
            life_bulk.append(numpy.ma.array(life, mask=mask))
        return numpy.hstack(
            [self.calc_stats_(numpy.ma.stack(bd2_bulk, axis=1)), self.calc_stats_(numpy.ma.stack(life_bulk, axis=1))]
        )


class DiagramVectorizer(cvtda.utils.FeatureExtractorBase):
    """
    Filters the persistence diagrams (removes a given fraction of least persistent features) and
    transforms them into feature vectors with the following values:
    - Statistical features of betti curves. See :class:`BettiCurve` for details.
    - Statistical features of persistence landscapes. See :class:`Landscape` for details.
    - Statistical features of persistence silhouettes. See :class:`Silhouette` for details.
    - Statistical features of heat kernels applied over diagrams. See :class:`HeatKernel` for details.
    - Statistical features of persistence images. See :class:`PersistenceImage` for details.
    - Persistence entropy of the diagrams. See :class:`Entropy` for details.
    - The number of points on the diagrams. See :class:`NumberOfPoints` for details.
    - Statistical features of simplex lifetimes. See :class:`Lifetime` for details.

    By the statistical features we imply the output of :func:`cvtda.utils.sequence2features`.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        filtering_percentile : ``float``, default = 10
            The fraction of least persistent features to remove during filtration.
        betti_curve : ``BettiCurve.Settings``
            Settings for the :class:`BettiCurve` calculation.
        landscape : ``Landscape.Settings``
            Settings for the :class:`Landscape` calculation.
        silhouette : ``Silhouette.Settings``
            Settings for the :class:`Silhouette` calculation.
        heat_kernel : ``HeatKernel.Settings``
            Settings for the :class:`HeatKernel` calculation.
        persistence_image : ``PersistenceImage.Settings``
            Settings for the :class:`PersistenceImage` calculation.
        entropy : ``Entropy.Settings``
            Settings for the :class:`Entropy` calculation.
        number_of_points : ``NumberOfPoints.Settings``
            Settings for the :class:`NumberOfPoints` calculation.
        lifetime : ``Lifetime.Settings``
            Settings for the :class:`Lifetime` calculation.
        """

        filtering_percentile: float = 10
        betti_curve: BettiCurve.Settings = BettiCurve.Settings()
        landscape: Landscape.Settings = Landscape.Settings()
        silhouette: Silhouette.Settings = Silhouette.Settings()
        heat_kernel: HeatKernel.Settings = HeatKernel.Settings()
        persistence_image: PersistenceImage.Settings = PersistenceImage.Settings()
        entropy: Entropy.Settings = Entropy.Settings()
        number_of_points: NumberOfPoints.Settings = NumberOfPoints.Settings()
        lifetime: Lifetime.Settings = Lifetime.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(
            betti_curve=BettiCurve.PRESETS.full,
            landscape=Landscape.PRESETS.full,
            silhouette=Silhouette.PRESETS.full,
            heat_kernel=HeatKernel.PRESETS.full,
            persistence_image=PersistenceImage.PRESETS.full,
            entropy=Entropy.PRESETS.full,
            number_of_points=NumberOfPoints.PRESETS.full,
            lifetime=Lifetime.PRESETS.full,
        ),
        reduced=Settings(
            betti_curve=BettiCurve.PRESETS.reduced,
            landscape=Landscape.PRESETS.reduced,
            silhouette=Silhouette.PRESETS.reduced,
            heat_kernel=HeatKernel.PRESETS.reduced,
            persistence_image=PersistenceImage.PRESETS.reduced,
            entropy=Entropy.PRESETS.reduced,
            number_of_points=NumberOfPoints.PRESETS.reduced,
            lifetime=Lifetime.PRESETS.reduced,
        ),
        quick=Settings(
            betti_curve=BettiCurve.PRESETS.quick,
            landscape=Landscape.PRESETS.quick,
            silhouette=Silhouette.PRESETS.quick,
            heat_kernel=HeatKernel.PRESETS.quick,
            persistence_image=PersistenceImage.PRESETS.quick,
            entropy=Entropy.PRESETS.quick,
            number_of_points=NumberOfPoints.PRESETS.quick,
            lifetime=Lifetime.PRESETS.quick,
        ),
    )

    EXTRACTOR_NAMES = [
        "betti",
        "landscape",
        "silhouette",
        "heat",
        "persistence_image",
        "entropy",
        "number_of_points",
        "lifetime",
    ]

    def __init__(self, n_jobs: int = -1, batch_size: int = None, settings: Settings = PRESETS.reduced):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.batch_size_ = batch_size

        self.filtering_percentile_ = settings.filtering_percentile
        self.extractors_: typing.List[cvtda.utils.FeatureExtractorBase] = [
            BettiCurve(settings.betti_curve),
            Landscape(settings.landscape),
            Silhouette(settings.silhouette),
            HeatKernel(settings.heat_kernel),
            PersistenceImage(settings.persistence_image),
            Entropy(settings.entropy),
            NumberOfPoints(settings.number_of_points),
            Lifetime(settings.lifetime),
        ]

    def feature_names(self) -> typing.List[str]:
        assert self.fitted_ is True, "fit() must be called before feature_names()"
        feature_names = []
        for extractor, name in zip(self.extractors_, DiagramVectorizer.EXTRACTOR_NAMES):
            feature_names.extend(self.nest_feature_names(name, extractor.feature_names()))
        return feature_names

    def fit(self, diagrams: numpy.ndarray):
        # Filtering
        self.filtering_epsilon_ = determine_filtering_epsilon_(diagrams, self.filtering_percentile_)
        self.filtering_ = gtda.diagrams.Filtering(epsilon=self.filtering_epsilon_).fit(diagrams)
        diagrams = self.filtering_.transform(diagrams)

        # Features
        for extractor in self.extractors_:
            extractor.fit(diagrams)

        cvtda.logging.logger().print("DiagramVectorizer: fitting complete")
        self.fitted_ = True
        return self

    def transform(self, diagrams: numpy.ndarray) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before transform()"
        batch_size = self.get_batch_size_(len(diagrams))
        loop = list(range(0, len(diagrams), batch_size))

        # Process diagrams in batches.
        # This appeared to be more efficient than using n_jobs in giotto-tda.
        def transform_batch(batch_start: numpy.ndarray) -> numpy.ndarray:
            return self.transform_batch_raw_(diagrams[batch_start : batch_start + batch_size])

        features = cvtda.utils.parallel(transform_batch, loop, return_as="generator", n_jobs=self.n_jobs_)
        collector = cvtda.logging.logger().pbar(features, total=len(loop), desc="DiagramVectorizer: batch")
        features = numpy.vstack(list(collector))
        assert features.shape == (len(diagrams), len(self.feature_names())), (
            f"{features.shape} != {(len(diagrams), len(self.feature_names()))}"
        )
        return features

    def get_batch_size_(self, num_objects: int):
        """
        Get efficient batch size for :meth:`transform`.

        Parameters
        ----------
        num_objects : ``int``
            The number of objects to process.

        Returns
        -------
        ``int``
            Batch size with empirically good balance between batching and parallelization.
        """

        if self.batch_size_ is not None:
            return self.batch_size_
        # Try to make batch size around 256, but to utilize all available CPU cores.
        n_jobs = joblib.effective_n_jobs(self.n_jobs_)
        batch_size = math.ceil(num_objects / n_jobs)
        return math.ceil(batch_size / math.ceil(batch_size / 256))

    def transform_batch_raw_(self, batch: numpy.ndarray) -> numpy.ndarray:
        """
        Filter and compute features for one batch of persistence diagrams.

        Parameters
        ----------
        batch : ``numpy.ndarray``
            (size `num_diagrams x num_points x 3`) A batch of persistence diagrams.

        Returns
        -------
        ``numpy.array``
            (size `num_diagrams x num_features`) The resulting features.
        """

        batch = self.filtering_.transform(batch)
        return numpy.hstack([extractor.transform(batch) for extractor in self.extractors_])



from .duplicates import DuplicateFeaturesRemover
from .FeatureExtractorBase import FeatureExtractorBase
from .image2pointcloud import image2pointcloud
from .rgb2gray import rgb2gray
from .rgb2hsv import rgb2hsv
from .sequence2features import sequence2features
from .set_random_seed import set_random_seed
from .spread_points import spread_points
from .parallel import parallel

__all__ = [
    "DuplicateFeaturesRemover",
    "FeatureExtractorBase",
    "image2pointcloud",
    "rgb2gray",
    "rgb2hsv",
    "sequence2features",
    "set_random_seed",
    "spread_points",
    "parallel",
]



import abc
import typing

T = typing.TypeVar("T")


class BaseDumper(abc.ABC, typing.Generic[T]):
    """
    Base class for dumping with context management support.
    """

    current_dumper = None

    def __enter__(self):
        self.__previous = BaseDumper.current_dumper
        BaseDumper.current_dumper = self
        return self

    def __exit__(self, *args):
        BaseDumper.current_dumper = self.__previous

    @abc.abstractmethod
    def execute(self, function: typing.Callable[[typing.Any], T], name: str, *function_args) -> T:
        """
        Gets the result from dump, if available, or executes the function and dumps the output.

        Parameters
        ----------
        function : ``Callable``
            The function to execute.
        name : ``str``
            The name of the dump.
        function_args : ``list[Any]``
            Extra arguments to pass to the function.

        Returns
        -------
        ``Any``
            The output of the function.
        """
        pass

    @abc.abstractmethod
    def save_dump(self, data: T, name: str):
        """
        Dumps data with the given name.

        Parameters
        ----------
        data : ``Any``
            The data to dump.
        name : ``str``
            The name of the dump.
        """
        pass

    @abc.abstractmethod
    def has_dump(self, name: str) -> bool:
        """
        Checks if the dump with a given name is available.

        Parameters
        ----------
        name : ``str``
            The name of the dump.

        Returns
        -------
        ``bool``
            True, if the dump is available. False otherwise.
        """
        pass

    def get_dump(self, name: str) -> T:
        """
        Ensures the dump with a given name is available and gets it.
        Throws :exc:`AssertionError` if the dump is not available.

        Parameters
        ----------
        name : ``str``
            The name of the dump.

        Returns
        -------
        ``Any``
            The contents of the dump.
        """
        assert self.has_dump(name), f"There is no dump at {name}"
        return self.get_dump_impl_(name)

    @abc.abstractmethod
    def get_dump_impl_(self, name: str) -> T:
        """
        Gets the dump with a given name without checking if it is available.

        Parameters
        ----------
        name : ``str``
            The name of the dump.

        Returns
        -------
        ``Any``
            The contents of the dump.
        """
        pass



import typing

import numpy
import torch
import torch.utils.data
import torchph.nn.slayer

import cvtda.utils
import cvtda.logging

from .device import default_device


def transform(diagram: torch.Tensor, dim: int):
    """
    Extracts a subdiagram corresponding to one homology dimnesion from the given persistence diagram
    and performs a rotation of the persistence diagrams into birth-persistence coordinates.
    See :class:`torchph.nn.slayer.LogStretchedBirthLifeTimeCoordinateTransform` for details.

    Parameters
    ----------
    diagram : ``torch.Tensor``
        (size `num_points x 3`) A persistence diagram.
    dim : ``int``
        Homology dimension to extract.

    Returns
    -------
    ``torch.Tensor``
        (size `num_points x 3`) Resulting subdiagram.
    """
    dim_filter = diagram[:, 2] == dim
    non_degenerate_filter = diagram[:, 0] < diagram[:, 1]
    rotation = torchph.nn.slayer.LogStretchedBirthLifeTimeCoordinateTransform(0.01)
    return rotation(diagram[dim_filter & non_degenerate_filter][:, 0:2])


def process_diagram(diags: torch.Tensor):
    """
    Preprocess a set of diagrams for convenient use with :mod:`torchph`.
    - Partitions the diagrams by homology dimensions.
    - Performs a rotation into birth-persistence coordinates. See :func:`transform` for details.
    - Prepares the diagrams for use in :mod:`torchph`. See :func:`torchph.nn.slayer.prepare_batch` for details.

    Parameters
    ----------
    diags : ``torch.Tensor``
        (size `num_diagrams x num_points x 3`) A batch of persistence diagrams.

    Returns
    -------
    ``tuple[torch.Tensor, torch.Tensor]``
        1. (size `num_diagrams x num_points x 3`) Preprocessed diagrams.
        2. (size `num_diagrams x num_points x 3`) Mask for non-dummy points on persistence diagrams.
        See :func:`torchph.nn.slayer.prepare_batch` for details.
    """
    diagrams, non_dummy_points = [], []
    for dim in diags[:, :, 2].unique(sorted=True):
        diags_dim = [transform(diag, dim) for diag in diags]
        processed = torchph.nn.slayer.prepare_batch(diags_dim)
        diagrams.append(processed[0].cpu())
        non_dummy_points.append(processed[1].cpu())
    return diagrams, non_dummy_points


class Dataset(torch.utils.data.Dataset):
    """
    A single structure representing the dataset with topological information.

    Attributes
    ----------
    images : ``torch.Tensor``
        (size `num_items x num_channels x width x height`) Raw images.
    labels : ``torch.Tensor``
        (size `num_items x ...`) The target variable.
        May be of any shape, if the first dimension is the number of items.
    features : ``torch.Tensor``
        (size `num_items x num_features`) Topological features for each image.
    raw_diagrams : ``list[numpy.ndarray]``, optional
        (size `n_items x n_diagrams x n_points x 3`) Raw persistence diagrams, if provided.
    diagrams : ``list[torch.Tensor]``, optional
        (size `n_items x n_diagrams x n_points x 3`) Processed persistence diagrams as single tensors.
    non_dummy_points : ``list[torch.Tensor]``, optional
        (size `n_items x n_diagrams x n_points`) Mask for non-dummy points on persistence diagrams.
        See :func:`transform` for details.
    """

    def __init__(
        self,
        images: numpy.ndarray,
        diagrams: typing.Optional[typing.List[numpy.ndarray]],  # n_items x n_diagrams x n_points x 3
        features: numpy.ndarray,
        labels: typing.Optional[numpy.ndarray],
        n_jobs: int = -1,
        device: torch.device = default_device,
    ):
        self.n_jobs_ = n_jobs
        self.device_ = device

        # Save images in a torch-compatible format.
        if images is not None:
            self.images = torch.tensor(images, dtype=torch.float32)
            if len(self.images.shape) == 4:
                self.images = self.images.permute((0, 3, 1, 2))
            else:
                assert len(self.images.shape) == 3
                self.images = self.images.unsqueeze(1)

        # Save everything else.
        if labels is not None:
            self.labels = torch.tensor(labels, dtype=torch.long)
        self.features = torch.tensor(features, dtype=torch.float32)
        self.raw_diagrams = diagrams

        if diagrams is None:
            return

        # Reorder the diagrams
        diagrams = [
            torch.tensor(numpy.array([item[num_diagram] for item in diagrams]), dtype=torch.float32)
            for num_diagram in range(len(diagrams[0]))
        ]

        # Process diagrams.
        pbar = cvtda.logging.logger().pbar(diagrams, desc="Dataset: processing diagrams")
        diagrams = cvtda.utils.parallel(process_diagram, pbar, n_jobs=self.n_jobs_)

        # Store
        self.diagrams, self.non_dummy_points = [], []
        for diag, ndp in diagrams:
            self.diagrams.extend(diag)
            self.non_dummy_points.extend(ndp)

        cvtda.logging.logger().print(
            f"Constructed a dataset of {len(self.images)} images of shape {self.images[0].shape} "
            + f"with {len(self.diagrams)} diagrams and {self.features.shape[1]} features"
        )

    def __len__(self):
        return len(self.images)

    def get_labels(self, idxs):
        """
        Gets the target variable on the required torch device for a batch of elements.

        Parameters
        ----------
        idxs : ``Iterable[int]``
            (size `batch_size`) The indexes of objects in the batch.

        Returns
        -------
        ``torch.Tensor``
            (size `batch_size x ...`) The target variable for the requested objects.
        """
        return self.labels[idxs].to(self.device_)

    def get_diagrams(self, idxs):
        """
        Gets the processed persistence diagrams on the required torch device for a batch of elements.

        Parameters
        ----------
        idxs : ``Iterable[int]``
            (size `batch_size`) The indexes of objects in the batch.

        Returns
        -------
        ``tuple[list[torch.Tensor], list[torch.Tensor]]``
            1. (size `batch_size x num_diagrams x num_points x 3`) Preprocessed diagrams.
            2. (size `batch_size x num_diagrams x num_points x 3`) Mask for non-dummy points on persistence diagrams.
        """
        output = []
        for diag, ndp in zip(self.diagrams, self.non_dummy_points):
            output.append(diag[idxs].to(self.device_))
            output.append(ndp[idxs].to(self.device_))
        return output



import numpy


def aggregate_(diff: numpy.ndarray) -> float:
    axes = tuple(set(range(len(diff.shape))) - set([0]))
    return numpy.sum(diff, axis=axes).mean()


def estimate_quality(decoded: numpy.ndarray, original: numpy.ndarray) -> dict:
    """
    Quality metrics for compression task.

    Parameters
    ----------
    decoded : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Decoded images.
    original : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Original images.

    Returns
    -------
    ``dict``
        1. MAE: mean absolute error between original and decoded images.
        2. MSE: mean squared error between original and decoded images.
    """
    decoded = numpy.squeeze(decoded)
    original = numpy.squeeze(original)
    return {"MAE": aggregate_(numpy.abs(original - decoded)), "MSE": aggregate_((original - decoded) ** 2)}



import abc
import typing

T = typing.TypeVar("T")


class BaseLogger(abc.ABC):
    """
    Base class for loggers with context management support.
    """

    current_logger = None

    def __enter__(self):
        self.__previous = BaseLogger.current_logger
        BaseLogger.current_logger = self
        return self

    def __exit__(self, *args):
        BaseLogger.current_logger = self.__previous

    @abc.abstractmethod
    def verbosity(self) -> int:
        """
        Gets the verbosity level of the logger.

        Returns
        -------
        ``int``
            The verbosity level.
        """
        pass

    @abc.abstractmethod
    def print(self, data: T, *args) -> None:
        """
        Prints data and args using the logger.

        Parameters
        ----------
        data : ``Any``
            The data to print.
        args : ``Any``
            Extra arguments to use for printing.
        """
        pass

    @abc.abstractmethod
    def pbar(
        self, data: typing.Iterable[T], total: typing.Optional[int] = None, desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        """
        Wraps an iterable in a progress bar that acts exactly
        like the original iterable, but prints a dynamically updating
        progressbar every time a value is requested.

        Parameters
        ----------
        data : ``Iterable``
            The original iterable.
        total : ``int``, optional
            The total number of elements in the iterable if len() is not available.
        desc : ``str``, optional
            The description to print with the progress bar.
        """
        pass

    @abc.abstractmethod
    def zip(self, *iterables, desc: typing.Optional[str] = None):
        """
        Equivalent of builtin `zip` with a progress bar printed by the logger.

        Parameters
        ----------
        iterables : ``list[Iterable]``
            The iterables to zip.
        desc : ``str``, optional
            The description to print with the progress bar.
        """
        pass

    @abc.abstractmethod
    def set_pbar_postfix(self, pbar: typing.Any, data: dict):
        """
        Sets a progress bar postfix (additional stats) with automatic formatting based on datatype.

        Parameters
        ----------
        pbar : ``Any``
            The progress bar instance.
        data : ``dict``
            The data to print with the progress bar.
        """
        pass
